#! perl
# generate a random polarium puzzle

use POSIX;

$maxsiz=8;
$minsiz=5;
$minflipped=.35; # 35 %
$maxflipped=.9; # 90 %
$solid = 0;

$maxrow = &randrange($minsiz, $maxsiz);
$maxcol = &randrange($minsiz, $maxsiz);

$file = (shift || "polarium");

$tstamp= POSIX::strftime("%Y-%m-%d-%H%M%S", localtime());
open (PUZ, ">>$file.txt");
open (PATH, ">>$file.path.txt");
print PUZ $tstamp, "\n";
print PATH $tstamp, "\n";

#set up board
for($row=1; $row<=$maxrow; ++$row) {
    for($col=1; $col<=$maxcol; ++$col) {
	$board[$row][$col] = 1;
    }
}
#set up border
for($row=0; $row<=$maxrow+1; ++$row) {
    $board[$row][0] = 0;
    $board[$row][$maxcol+1]=0;
}
for($col=1; $col<=$maxcol; ++$col) {
    $board[0][$col] = 0;
    $board[$maxrow+1][$col] = 0;
}
#set up path
for($row=0; $row<$maxrow+2; ++$row) {
    for($col=0; $col < $maxcol + 2; ++$col) {
	$path[$row][$col] = 0;
    }
}
unless($solid) { #flip random rows
    for($row=1; $row<=$maxrow; ++$row) {
	if(rand() < .5) {
	    for($col=1;$col<=$maxcol;++$col) {
		$board[$row][$col] = 2;
	    }
	}
    }
}

#Make a random walk
$numtoflip = &randrange($maxrow * $maxcol * $minflipped,  $maxrow * $maxcol * $maxflipped);
for( $flipped = 1,
     $at[0] = &randrange(1, $maxrow),
     $at[1] = &randrange(1, $maxcol);
     $flipped < $numtoflip;
     ) {
    $to = &nextdir;
    if($flipped == 1) {
	$path[$at[0]][$at[1]] = "S";
	@start = @at;
	if($board[$at[0]][$at[1]] == 2) {
	    $board[$at[0]][$at[1]] = 1; ++$flipped;
	} elsif ($board[$at[0]][$at[1]] == 1) {
	    $board[$at[0]][$at[1]] = 2; ++$flipped;
	}
    } else {
	$path[$at[0]][$at[1]] = &pathmark(@$to);
	#print &pathmark(@$to);
    }
    @nxt = ($at[0]+$to->[0], $at[1]+$to->[1]);
    @at = @nxt;
    if($board[$at[0]][$at[1]] == 2) {
	$board[$at[0]][$at[1]] = 1; ++$flipped; @lastflip=@at;
    } elsif ($board[$at[0]][$at[1]] == 1) {
	$board[$at[0]][$at[1]] = 2; ++$flipped; @lastflip=@at;
    }
    #print "direction ", $to->[0], " ", $to->[1], "\n";
    last if ($to->[0] == 0 && $to->[1] == 0); #trapped
}    
@end = @at;
$path[$at[0]][$at[1]] = "E";

#print puzzle
for($row = 1; $row<=$maxrow; ++$row) {
    for($col = 1; $col <= $maxcol; ++$col) {
	if($board[$row][$col] == "2") {
	    print PUZ "@ ";
	} else {
	    print PUZ "- ";
	}
    }
    print PUZ "\n";
}

#print path
for($row = 0; $row<=$maxrow+1; ++$row) {
    for($col = 0; $col <= $maxcol+1; ++$col) {
	print PATH $path[$row][$col], " ";
    }
    print PATH  "\n";
}
#print polarium password
if($maxcol <= 8 && $maxrow <= 8) {
    &polariumpass;
}
&stroqpass;
print PUZ "-"x30, "\n";
print PATH "-"x30, "\n";


sub randrange {
    my $min = shift || ($min=0);
    my $max = shift || ($max=$min+1);
    return int(rand($max - $min+1)) + $min;
}

sub nextdir {
    my @dirs = ();
    if($at[0]>0 && $path[$at[0]-1][$at[1]] eq "0" ) { #up
	push(@dirs, [-1, 0]);
    }
    if($at<$maxrow+1 && $path[$at[0]+1][$at[1]] eq "0") { #down
	push(@dirs, [1, 0]);
    }
    if($at[1]<$maxcol+1 && $path[$at[0]][$at[1]+1] eq "0") { #right
	push(@dirs, [0, 1]);
    }
    if($at[1]>0 && $path[$at[0]][$at[1]-1] eq "0") { #left
	push(@dirs, [0, -1]);
    }
    if(scalar(@dirs) == 0) {
	return( [0,0]);
    }
    if($#dirs == 0) {return $dirs[0];}
    return($dirs[rand($#dirs)]);
    #$path[$at[0]][$at[1]] = 1;
}

sub pathmark {
    my $r = shift;
    my $c = shift;
    return ("^") if $r < 0;
    return ("v") if $r > 0;
    return ("<") if $c < 0;
    return (">") if $c > 0;
    return ("?");
}

sub polariumpass {
    my $bits = "";
    my ($row, $col);
    for ($row = 1; $row < 9; ++$row) {
	for ($col = 8; $col >= 1; --$col) {
	    if($board[$row][$col] == 2) {
		$bits = $bits . "1";
	    } elsif ($board[$row][$col] == 1) {
		$bits = $bits . "0";
	    } else {
		$bits = $bits . "0";
	    }
	}
    }
    #print $bits, "\n";
    @hex = unpack("H2H2H2H2H2H2H2H2", pack("B32B32", $bits, substr($bits,32)));
    $hex[11] = "00"; $hex[10] = $maxrow . $maxcol; $hex[9] = $start[0]. $start[1]; $hex[8] = $end[0].$end[1];
    $chksm = 0;
	for $byteval (@hex[0,1,2,3,4,5,6,7,8,9,10]) {
	   # if($byteval eq "00") {
	#	$byteval = "AA";
	#    }
	    $chksm += hex($byteval);
	    #print "$chksm $byteval ";
	}
	$chksm = $chksm & 255;
    $hex[11] = unpack("H2", pack("S", $chksm));
    #print join(" ", @hex), "\n";
    for $num (0..2) {
	$line[$num] = 16777216 * hex($hex[$num*4+3]) + 65536 * hex($hex[$num*4+2]) + 256 * hex($hex[$num*4+1])
	    + hex($hex[$num*4]);
	$line[$num] = sprintf("%010u", $line[$num]);
	$line[$num] = join("", reverse(split("", $line[$num])));
    }
	#unpack("L", (pack("H2H2H2H2"), @hex[3,2,1,0]));
    print PUZ "Polarium: ", $line[0], "\n"," "x10, $line[1], "\n"," "x10, $line[2], "\n";
}

sub stroqpass {

    if($maxcol > $maxrow) {
	$bitw = int(log($maxcol) / log(2) + 1);
    } else {
	$bitw = int(log($maxrow) / log(2) + 1);
    }
    
    #print $bitw, "\n";
    $bits = substr(unpack("B8", pack("S", $bitw)), -6)
	. substr(unpack("B8", pack("S", $maxcol)),-$bitw)
	. substr(unpack("B8", pack("S", $maxrow)),-$bitw)
	. substr(unpack("B8", pack("S", $start[1])),-$bitw)
	. substr(unpack("B8", pack("S", $start[0])),-$bitw)
	. substr(unpack("B8", pack("S", $lastflip[1])),-$bitw)
	. substr(unpack("B8", pack("S", $lastflip[0])),-$bitw);

    for ($row = 1; $row <= $maxrow; ++$row) {
	for($col = 1; $col <= $maxcol; ++$col) {
	    if($board[$row][$col] == 2) {
		$bits .= "1";
	    } else {
		$bits .= "0";
	    }
	}
    }
    
# pad with '0' to make multiple of 8
    $bits .= substr("00000000", 0,  -length($bits) % 8);
    $output = unpack("H*", pack("B*", $bits));
    
    #print $bits, "\n", $output, "\n";
    $output =~ tr/0-9a-f/A-P/;
    print PUZ "Stroq: ", $output, "\n";
    
}
